LIBRARY HawkNL


EXPORTS

nlAcceptConnection
nlAddrCompare
nlAddrToString
nlClear
nlClearSocketStat
nlClose
nlCondBroadcast
nlCondDestroy
nlCondInit
nlCondSignal
nlCondWait
nlConnect
nlDisable
nlEnable
nlGetAddrFromName
nlGetAddrFromNameAsync
nlGetAllLocalAddr
nlGetBoolean
nlGetCRC16
nlGetCRC32
nlGetError
nlGetErrorStr
nlGetInteger
nlGetLocalAddr
nlGetNameFromAddr
nlGetNameFromAddrAsync
nlGetPortFromAddr
nlGetRemoteAddr
nlGetSocketStat
nlGetString
nlGetSystemError
nlGetSystemErrorStr
nlGroupAddSocket
nlGroupCreate
nlGroupDeleteSocket
nlGroupDestroy
nlGroupGetSockets
nlHint
nlInit
nlListen
nlMutexDestroy
nlMutexInit
nlMutexLock
nlMutexUnlock
nlOpen
nlPollGroup
nlRead
nlSelectNetwork
nlSetAddrPort
nlSetLocalAddr
nlSetRemoteAddr
nlShutdown
nlStringToAddr
nlSwapd
nlSwapf
nlSwapl
nlSwaps
nlThreadCreate
nlThreadJoin
nlThreadSleep
nlThreadYield
nlTime
nlWrite
